/*

	CRWCLS.PRG   29 Apr 94
	
	Author: Daniel P. Evans, dba DataCraft
			(206) 841-0819
			CIS:		73717,463
			Internet:	devans@eskimo.com
	
	Crystal Report Writer class-- encapsulates all of the CRW functionality
		via CRW's PrintEngine using Clip4Win & Classy

	Usage:
		Put CRW():Initclass() and CRW():Killclass() in the INIT and
		EXIT routines for an app.  The ::Initclass() sets up the
		Crystal printengine DLL and opens the engine while ::Killclass
		closes both the engine and the DLL.

		To run a report, simply
		
		 	CRW():New( <ReportName>,<WindowTitle> ):Run()

		<ReportName> should include the full path the report if not in
		the same directory.  <WindowTitle> is really only necessary if
		the report is being sent to the window.
		
		The several Formula and Selection methods can be used to send
		selection criteria or runtime variable (Crystal "formula") values
		to the report.
		
	Gaps/Problems:
	
		Haven't been able to get access the window handle for the
		report's window ( see ::hWnd in ::GetDestination ) so have
		disabled it.
		
		::GetDestination() does not yet have a dialog box for getting
		the destination from the user at runtime.  This could probably
		even be done very simply with a call to MessageBox() although
		a full dialogbox() would be best.  Could also incorporate
		C4W's PrintDlg() to allow selection of specific printer.

		
	Last change:  DPE   8 Jun 94   11:22 am
*/

#define	WIN_WANT_LBS
#define	WIN_WANT_ALL
#define	WIN_WANT_EN

#include	"windows.ch"
#include	"setcaret.ch"
#include	"dialog.ch"

#include	"COMMON.CH"
#include	"SET.CH"
#include	"INKEY.CH"

#include	"CLASS(Y).CH"

// following are from Crystal's CRPE.H
#define PE_FT_RECORD              0
#define PE_FT_TABSEPARATED        1
#define PE_FT_TEXT                2
#define PE_FT_DIF                 3
#define PE_FT_CSV                 4
#define PE_FT_CHARSEPARATED       5
#define PE_FT_TABFORMATTED        6

#define PE_SF_MAX_NAME_LENGTH 50

#define PE_SF_DESCENDING 0
#define PE_SF_ASCENDING  1

#define PE_ERR_NOERROR              0

#define PE_ERR_NOTENOUGHMEMORY      500
#define PE_ERR_INVALIDJOBNO         501
#define PE_ERR_INVALIDHANDLE        502
#define PE_ERR_STRINGTOOLONG        503
#define PE_ERR_NOSUCHREPORT         504
#define PE_ERR_NODESTINATION        505
#define PE_ERR_BADFILENUMBER        506
#define PE_ERR_BADFILENAME          507
#define PE_ERR_BADFIELDNUMBER       508
#define PE_ERR_BADFIELDNAME         509
#define PE_ERR_BADFORMULANAME       510
#define PE_ERR_BADSORTDIRECTION     511
#define PE_ERR_ENGINENOTOPEN        512
#define PE_ERR_INVALIDPRINTER       513
#define PE_ERR_PRINTFILEEXISTS      514
#define PE_ERR_BADFORMULATEXT       515
#define PE_ERR_BADGROUPSECTION      516
#define PE_ERR_ENGINEBUSY           517
#define PE_ERR_BADSECTION           518
#define PE_ERR_NOPRINTWINDOW        519
#define PE_ERR_JOBALREADYSTARTED    520
#define PE_ERR_BADSUMMARYFIELD      521
#define PE_ERR_NOTENOUGHSYSRES      522
#define PE_ERR_BADGROUPCONDITION    523
#define PE_ERR_JOBBUSY              524
#define PE_ERR_BADREPORTFILE        525
#define PE_ERR_NODEFAULTPRINTER     526
#define PE_ERR_SQLSERVERERROR       527
#define PE_ERR_BADLINENUMBER        528
#define PE_ERR_DISKFULL             529
#define PE_ERR_FILEERROR            530
#define PE_ERR_INCORRECTPASSWORD    531
#define PE_ERR_BADDATABASEDLL       532
#define PE_ERR_BADDATABASEFILE      533
#define PE_ERR_ERRORINDATABASEDLL   534
#define PE_ERR_DATABASESESSION      535
#define PE_ERR_DATABASELOGON        536
#define PE_ERR_DATABASELOCATION     537
#define PE_ERR_BADSTRUCTSIZE        538
#define PE_ERR_BADDATE              539
#define PE_ERR_BADEXPORTDLL         540
#define PE_ERR_ERRORINEXPORTDLL     541
#define PE_ERR_PREVATFIRSTPAGE      542
#define PE_ERR_NEXTATLASTPAGE       543
#define PE_ERR_CANNOTACCESSREPORT   544
#define PE_ERR_USERCANCELLED        545
#define PE_ERR_OLE2NOTLOADED        546
#define PE_ERR_BADCROSSTABGROUP     547
#define PE_ERR_NOCTSUMMARIZEDFIELD  548
#define PE_ERR_DESTINATIONNOTEXPORT 549

#define PE_ERR_NOTIMPLEMENTED       999

/* ----------------------------------------------------------------- */
CREATE CLASS CRW

protected:

	class var	hLibrary
	class var	lEngineOpen

	class method	OpenLibrary
	class method	OpenEngine

	// need to shut things down
	class method	CloseEngine
	class method	CloseLibrary

	var		SelectionFormula
	var		Formulas
	
	var		PrintDriver
	var		PrinterName
	var		Port
	var		Copies

	method	GetError

export:

	class message	Library	method	GetLibrary		// access name
	class method	_Library						// change name
	
	class message	EngineOpen method GetEngOpen	// access status
	class method	_EngineOpen						// change status

	
	class method	initclass
	class method	killclass

	var		ReportName								// full name w/ path
	var		PrintJob								// provided by PE
	var		Destination								// coded or ::GetDest
	var		Title									// for preview window
	var		Selection								// formula for filter

	var		OutputFile								// file name, w/ path
	var		FileOptions								// structure for fileopt

//	var		hWnd									// report window handle
	
	method	init

	method	Run

	method  OpenPrintJob                            // call PEOpenPrintJob
	method	GetDestination							// get dest & set ::Destination
	method	SetDestination							// call PE based on ::Destination
	method	SetSelection							// set filter
	method  SetSelFormula                           // call PE to set it
	method	GetFormula								// get a runtime variable
	method	SetFormula								// set the formulas
	
ENDCLASS
/* ----------------------------------------------------------------- */
METHOD initclass

::OpenLibrary()

::OpenEngine()

RETURN (self)
/* ----------------------------------------------------------------- */
METHOD killclass

::CloseEngine()

::CloseLibrary()

RETURN (self)
/* ----------------------------------------------------------------- */
METHOD OpenLibrary()
/*
	open the Crystal DLL
*/

::hLibrary	:=	LoadLibrary( "CRPE.DLL" )

RETURN (self)
/* ----------------------------------------------------------------- */
METHOD OpenEngine()
/*
	open the Crystal Engine
*/
local	cProc

if ::hLibrary != NIL

	cProc			:=	GetProcAddress( ::hLibrary, "PEOpenEngine",;
							"PASCAL", "log", "void" )
					
	if !(::lEngineOpen	:=	CallDLL( cProc ))
		// problem so set library to NIL
		::hLibrary	:=	NIL
	endif

endif

// give an error here if ::hLibraray == NIL
if ::hLibrary	==	NIL
	MessageBox( SelectWindow(),;
		"The report routine library could not be opened."+CRLF+;
		"You will not be able to run most reports."+CRLF+;
		"Please notify your system administrator to have this corrected.",;
		"No Report Library", MB_OK+MB_ICONEXCLAMATION )
endif

RETURN (self)
/* ----------------------------------------------------------------- */
METHOD GetLibrary()					// message is Library()

RETURN (::hLibrary)
/* ----------------------------------------------------------------- */
METHOD _Library( hLibrary )

::hLibrary	:=	hLibrary

RETURN (hLibrary)
/* ----------------------------------------------------------------- */
METHOD GetEngOpen()					// message is EngineOpen()

RETURN (::lEngineOpen)
/* ----------------------------------------------------------------- */
METHOD _EngineOpen( lStatus )

::lEngineOpen	:=	lStatus

RETURN (lStatus)
/* ----------------------------------------------------------------- */
METHOD CloseEngine()

local	cProc

if ::hLibrary != NIL

	cProc			:=	GetProcAddress( ::hLibrary, "PECloseEngine",;
							"PASCAL", "void", "void" )

	CallDLL( cProc )
	
	::lEngineOpen	:=	FALSE

endif

RETURN (self)
/* ----------------------------------------------------------------- */
METHOD CloseLibrary()

FreeLibrary( ::hLibrary )

::hLibrary	:=	NIL

RETURN (self)
/* ----------------------------------------------------------------- */
METHOD init( cReport, cTitle ),()

default cTitle to cReport

::ReportName		:=	alltrim(cReport)+" "
::Title				:=	alltrim(cTitle)+" "

::Destination		:=	"SCREEN"					// by default
::SelectionFormula	:=	NIL
::Formulas			:=	{}							// runtime variables

::Copies			:=	1

::OutputFile		:=	""
::FileOptions		:=	""

// ::hWnd				:=	NIL

RETURN (self)
/* ----------------------------------------------------------------- */
METHOD Run()
/*
	run ::ReportName
*/
local	cProc,			;
		nError

if ::class:hLibrary != NIL;
	.and. ::ReportName != NIL

    ::OpenPrintJob()
    
	if ::PrintJob != 0

		::SetDestination()

		// setup any runtime variables		
		::SetFormula()

		// now set up the selection formula
		::SetSelFormula()

		// run the thing
		cProc	:=	GetProcAddress( ::class:hLibrary, "PEStartPrintJob",;
						"Pascal", "log", "int,log" )
		CallDLL( cProc, ::PrintJob, TRUE )

		// now close the print job
		cProc	:=	GetProcAddress( ::class:hLibrary, "PEClosePrintJob",;
						"Pascal", "void", "int" )
		CallDLL( cProc, ::PrintJob )
		
	else
		// then do some error recovery
		::GetError( "Open Print Job" )

	endif
	
endif

RETURN (self)
/* ----------------------------------------------------------------- */
METHOD OpenPrintJob()
/*
    call PE func PEOpenPrintJob()
*/
local   cProc
  
cProc		:=	GetProcAddress( ::class:hLibrary, "PEOpenPrintJob",;
					"Pascal", "int", "string" )
						
::PrintJob	:=	CallDLL( cProc, ::ReportName )
	

RETURN (self)
/* ----------------------------------------------------------------- */
METHOD SetDestination()
/*
	call appropriate PE functions based on ::Destination
	
*/
local	cProc,			;
		lOut

do case

    case ::Destination	==	"PRINTER"
    
    	cProc	:=	GetProcAddress( ::class:hLibrary, "PEOutputToPrinter",;
    					"Pascal", "log", "int,int" )
    	
    	lOut	:=	CallDLL( cProc, ::PrintJob, ::Copies )
    
    case ::Destination	==	"FILE"
    
    	cProc	:=	GetProcAddress( ::class:hLibrary, "PEOutputToFile",;
    					"Pascal", "log", "int,string,int,string" )
    					
    	lOut	:=	CallDLL( cProc, ::PrintJob, ::OutputFile,;
    					PE_FT_TABFORMATTED, ::FileOptions )

	otherwise
		// default is screen
        cProc	:=	GetProcAddress( ::class:hLibrary, "PEOutputToWindow",;
				        "Pascal", "log", "int,string,int,int,int,int,long,int" )

        lOut	:=	CallDLL( cProc, ::PrintJob, ::Title, 0, 0,;
				        CW_USEDEFAULT, CW_USEDEFAULT, 0, 0 )
				
        cProc	:=	GetProcAddress( ::class:hLibrary, "PEShowPrintControls",;
				        "Pascal", "log", "int,log" )
				
        CallDLL( cProc, ::PrintJob, TRUE )

/*  this doesn't seem to work 21 May 94        
        cProc	:=	GetProcAddress( ::class:hLibrary, "PEGetWindowHandle",;
        				"Pascal", "int", "int" )
        				
        ::hWnd	:=	CallDLL( cProc, ::PrintJob )
*/
    
endcase


RETURN (self)
/* ----------------------------------------------------------------- */
METHOD SetSelection( cString )

::SelectionFormula	:=	alltrim(cString)+" "

RETURN (self)
/* ----------------------------------------------------------------- */
METHOD SetSelFormula()
/*
	set selectionformula if not empty
*/
local	cProc

if ::SelectionFormula != NIL
	cProc	:=	GetProcAddress( ::class:hLibrary,;
					"PESetSelectionFormula", "Pascal",;
					"log", "int, string" )
	if !( CallDLL( cProc, ::PrintJob, ::SelectionFormula ) )
		// error in selection formula
		::GetError( "Setting Selction Formula" )
	endif
endif

RETURN (self)
/* ----------------------------------------------------------------- */
METHOD GetDestination( cDest )
/*
	get destination from user & set ::Destination
	
	cDest is an optional param used to set ::Destination, if it is
		empty a dialog will be used to get the destination
	
	NOTE: need to also call printer selection routine
*/


if empty( cDest )
	// then run a dialog to get it from the user
	
	// if they cancelled, clear the reportname so the report won't run
	if FALSE
		::ReportName	:=	NIL
	endif
	
elseif cDest != "SCREEN" .and. cDest != "PRINTER";
	.and. cDest != "FILE"
	// oops-- mistake so just use screen
	cDest	:=	"SCREEN"

endif

::Destination	:=	"SCREEN"

RETURN (self)		
/* ----------------------------------------------------------------- */
METHOD GetFormula( cName, cString )
/*
	add cName and cString to array of forumlas (runtime variables)
*/

aadd( ::Formulas, { cName, asString(cString) } )

RETURN (self)
/* ----------------------------------------------------------------- */
METHOD SetFormula()
/*
	call the CRW procedure to set formulas
*/
local	cProc,		;
		lSet,		;
		i,			;
		n

if (n := len(::Formulas)) > 0

    cProc	:=	GetProcAddress( ::class:hLibrary, "PESetFormula",;
    				"Pascal", "log", "int,str,str" )

	for i := 1 to n    	

	    if !CallDLL( cProc, ::PrintJob, ::Formulas[i,1], ::Formulas[i,2] )
	    	::GetError( "Set Formula" )
	    	MessageBox( , "The value for "+asString(::Formulas[i,1])+CRLF+;
	    		"was wrong: "+CRLF+::Formulas[i,2] )
	    endif
	    
	next
	
endif

RETURN (self)
/* ----------------------------------------------------------------- */
METHOD GetError( cFunc )
/*
	get a CRW PE error & report it
*/
local	cProc,			;
		nError

cProc	:=	GetProcAddress( ::class:hLibrary, "PEGetErrorCode",;
				"Pascal", "int", "int" )
nError	:=	CallDLL( cProc, ::PrintJob )
		
MessageBox( SelectWindow(),;
	"There was a report error during"+CRLF+cFunc+CRLF+;
	"The Crystal Error number is: "+asString(nError),;
	"Report Error!", MB_OK+MB_ICONEXCLAMATION )

RETURN (self)

